searchData={"items":[{"type":"module","title":"OSCx","doc":"This is the main OSCx module.\n\nOSCx is an Elixir library for encoding and decoding Open Sound Control (OSC) messages and bundles.\n\nOSC is a flexible protocol that can be used for a wide variety of real-time tasks, such as controlling multimedia devices.\n\nNote that this library is deliberately minimilistic and no network transport or process logic is included.","ref":"OSCx.html"},{"type":"module","title":"Concepts, messages and bundles - OSCx","doc":"You can learn more about OSC and it's concepts and how to use this library by:\n- accessing the [README file](README.md)\n- taking the [Livebook tour](livebook/oscx_tour.livemd)\n- browsing the simple [Examples](examples.md)\n- reading the `OSCx.Message` or `OSCx.Bundle` module documentation\n- reading about OSC [arguments and types](arguments_and_types.md)\n\n#","ref":"OSCx.html#module-concepts-messages-and-bundles"},{"type":"module","title":"Example - OSCx","doc":"To encode a basic message:\n```\niex> my_message = %OSCx.Message{address: \"/my_synth/volume\", arguments: [0.25]}\n%OSCx.Message{address: \"/my_synth/volume\", arguments: [0.25]}\n\niex> binary_message = OSCx.encode(my_message)\n >\n```\nTo decode a binary OSC message:\n```\niex> binary_message =  >\n >\n\niex> decoded_message = OSCx.decode(binary_message)\n%OSCx.Message{address: \"/my_synth/volume\", arguments: [0.25]}\n```","ref":"OSCx.html#module-example"},{"type":"module","title":"TCP - OSCx","doc":"If using TCP as the transport mechanism, append the `:tcp` atom to the encode function:\n```\nOSCx.encode(my_message_or_bundle, :tcp)\n```\nThis frames the OSC packet with the message length in the header. This isn't required by UDP. See `encode/2` for more information.","ref":"OSCx.html#module-tcp"},{"type":"function","title":"OSCx.decode/1","doc":"Decodes a binary OSC Message or Bundle.","ref":"OSCx.html#decode/1"},{"type":"function","title":"Example - OSCx.decode/1","doc":"```\niex> binary_msg =  >\n >\n\niex> binary_msg |> OSCx.decode()\n%OSCx.Message{address: \"/status\", arguments: [0]}\n```","ref":"OSCx.html#decode/1-example"},{"type":"function","title":"OSCx.encode/1","doc":"","ref":"OSCx.html#encode/1"},{"type":"function","title":"OSCx.encode/2","doc":"Encodes an `%OSCx.Message{}` or `%OSCx.Bundle{}` struct as OSC binary data.\n\nTakes as `%OSCx.Message{}` or `%OSCx.Bundle{}` as the first parameter.","ref":"OSCx.html#encode/2"},{"type":"function","title":"Example - OSCx.encode/2","doc":"```\n# Encode a message\niex> %OSCx.Message{address: \"/status\", arguments: [1]} |> OSCx.encode()\n >\n\n# Encode a bundle with a message\niex> OSCx.Bundle.new(\n  elements: [OSCx.Message.new()],\n  time: %{seconds: 1, fraction: 100}\n  )\n  |> OSCx.encode()\n[\n   >,\n  {116,  >},\n  [ >]\n]\n```","ref":"OSCx.html#encode/2-example"},{"type":"function","title":"TCP - OSCx.encode/2","doc":"If using TCP as the transport mechansim, pass the `:tcp` atom as the second parameter:\n\n```\nOSCx.encode(my_message_or_bundle, :tcp)\n```\n\nThis will prefix the length (byte size) of the OSC binary data in 32-bits.\n\nThis is because in stream-based protocols such as TCP, the outgoing OSC packets are required to be framed using the packet's length in the header.\n\nThis isn't needed if using UDP.","ref":"OSCx.html#encode/2-tcp"},{"type":"module","title":"OSCx.Bundle","doc":"A module and struct for manipulating and representing OSC bundles.\n\nThe struct has two keys:\n- `time:` a map representing an OSC time tag, in the format `%{time: _, fraction: _}`\n- `elements:` an Elixir list of Messages or Bundles.\n\nThe two main functions are:\n- `encode/1` which takes an `%OSCx.Bundle{}` struct and encodes it to the OSC bundle format\n- `decode/1` which takes an OSC bundle recieved (e.g. via UDP) and decodes it into an `%OSCx.Bundle{}` struct.","ref":"OSCx.Bundle.html"},{"type":"module","title":"About OSC bundles - OSCx.Bundle","doc":"Bundles are a way of grouping OSC messages and even other OSC bundles together, so they can be received by the OSC server simultaneously.\n\n#","ref":"OSCx.Bundle.html#module-about-osc-bundles"},{"type":"module","title":"Structure of an OSC bundle - OSCx.Bundle","doc":"A bundle is made up of three parts:\n- Bundle identifer: which is the the string “#bundle”.\n- Time: a time tag which is a 64-bit time identifier. The first 32 bits specify the number of seconds since midnight on January 1, 1900, and the last 32 bits specify fractional parts of a second to a precision of about 200 picoseconds. This representation is used by Internet NTP timestamps.\n- Elements: the payload of the bundle, which can be any number of messages or bundles. Each of these are preceded by a 4-byte integer byte count.\n\n![OSC bundle diagram](assets/osc-bundle.png)\n\nCalling `OSCx.encode/1` or `OSCx.Bundle.encode/1` on an `%OSCx.Bundle{}` will create the OSC binary version matching the above.","ref":"OSCx.Bundle.html#module-structure-of-an-osc-bundle"},{"type":"module","title":"Creating a bundle - OSCx.Bundle","doc":"To create a bundle, you only need to specify the time tag and add the elements (which are `%OSCx.Messages{}` or other `%OSCx.Bundle{}` structs) to be included within the bundle:\n```\niex> my_bundle =\n  %OSCx.Bundle{\n    time: %{seconds: 1, fraction: 0},\n    elements: [%OSCx.Message{address: \"/\", arguments: []}]\n  }\n```\nOr you can use `OSCx.Bundle.new/1` to build the struct.","ref":"OSCx.Bundle.html#module-creating-a-bundle"},{"type":"module","title":"Sending the bundle - OSCx.Bundle","doc":"The bundle is sent just like an `OSCx.Message`, for example, using UDP:\n```\n# IP or host and port number for the UDP connection\nip_address = '127.0.0.1' # This could be changed to named address, like 'localhost'\nport_num = 8000 # In this example, this is the default port used by [Protokol](https://hexler.net/protokol)\n\n# Open a port\n{:ok, port} = :gen_udp.open(0, [:binary, {:active, true}])\n\n# Encode the message\nmy_bundle =\n  %OSCx.Bundle{\n    time: %{seconds: 1, fraction: 0},\n    elements: [\n      %OSCx.Message{address: \"/msg/one\", arguments: [1]},\n      %OSCx.Message{address: \"/msg/two\", arguments: [2]},\n      %OSCx.Message{address: \"/msg/three\", arguments: [3]}\n      ]\n  } |> OSCx.encode()\n\n# Send message\n:gen_udp.send(port, ip_address, port_num, my_bundle)\n```","ref":"OSCx.Bundle.html#module-sending-the-bundle"},{"type":"function","title":"OSCx.Bundle.decode/1","doc":"Decodes an OSC binary bundle into an `%OSCx.Bundle{}` struct.","ref":"OSCx.Bundle.html#decode/1"},{"type":"function","title":"Example - OSCx.Bundle.decode/1","doc":"```\niex> alias OSCx.{Bundle, Message}\n[OSCx.Bundle, OSCx.Message]\n\niex> binary_osc_bundle =\n  Bundle.new(\n    time: %{seconds: 1, fraction: 1},\n    elements: [\n      Message.new(address: \"/synth/play\", arguments: [440.0]),\n      Message.new(address: \"/synth/stop\", arguments: [440.0]),\n      Bundle.new(\n        time: %{seconds: 1, fraction: 2},\n        elements: [ Message.new(address: \"/inner_bundle/message\", arguments: [\"yes, they can be nested!\"]) ]\n      )\n    ]\n  ) |> OSCx.encode()\n >\n\n# Decode the binary OSC message above, back into a struct\niex> binary_osc_bundle |> OSCx.decode()\n%OSCx.Bundle{\n  time: %{seconds: 1, fraction: 1},\n  elements: [\n    %OSCx.Message{address: \"/synth/play\", arguments: [440.0]},\n    %OSCx.Message{address: \"/synth/stop\", arguments: [440.0]},\n    %OSCx.Bundle{\n      time: %{seconds: 1, fraction: 2},\n      elements: [\n        %OSCx.Message{\n          address: \"/inner_bundle/message\",\n          arguments: [\"yes, they can be nested!\"]\n        }\n      ]\n    }\n  ]\n}\n```","ref":"OSCx.Bundle.html#decode/1-example"},{"type":"function","title":"OSCx.Bundle.encode/1","doc":"Encodes an `%OSCx.Bundle{}` struct to an OSC Bundle binary.\n\nTakes a `%OSCx.Bundle{}` struct as it's first parameter.\n\nIn practice, just use `OSCx.encode/1` instead as it will encode Bundle or Message types.","ref":"OSCx.Bundle.html#encode/1"},{"type":"function","title":"Example - OSCx.Bundle.encode/1","doc":"```\niex> my_bundle =\n  Bundle.new(\n    time: %{seconds: 1, fraction: 1},\n    elements: [\n      Message.new(address: \"/synth/play\", arguments: [440.0]),\n      Message.new(address: \"/synth/stop\", arguments: [440.0])\n    ]\n  )\n%OSCx.Bundle{\n  time: %{seconds: 1, fraction: 1},\n  elements: [\n    %OSCx.Message{address: \"/synth/play\", arguments: [440.0]},\n    %OSCx.Message{address: \"/synth/stop\", arguments: [440.0]}\n  ]\n}\n\niex> my_bundle |> OSCx.Bundle.encode()\n >\n```","ref":"OSCx.Bundle.html#encode/1-example"},{"type":"function","title":"OSCx.Bundle.new/1","doc":"Convenience for creating a new `%OSCx.Bundle{}` struct.\n\nOptionally takes a keyword list with the following keys:\n- `time:` which can be used to set the OSC time tag map, e.g. `%{seconds: 1, fraction: 1}`\n- `elements:` which is a list of elements to include in the bundle. Elements can include zero, 1 or more `%OSCx.Message{}` or `%OSCx.Bundle{}` structs.","ref":"OSCx.Bundle.html#new/1"},{"type":"function","title":"Example - OSCx.Bundle.new/1","doc":"```\nBundle.new(\n    time: %{seconds: 1, fraction: 1},\n    elements: [\n      Message.new(address: \"/synth/play\", arguments: [440.0]),\n      Message.new(address: \"/synth/stop\", arguments: [440.0]),\n    ]\n  )\n```","ref":"OSCx.Bundle.html#new/1-example"},{"type":"module","title":"OSCx.Message","doc":"A module and struct for manipulating and representing OSC messages.\n\nThe struct has two keys:\n- `address:` representing the OSC address. Defaults to the root address of `\"/\"`\n- `arguments:` an Elixir list of arguments. Defults to an empty list `[]`.\n\nThe arguments list can include a range of values, such as:\n- Integer, e.g. `1`\n- Float, e.g. `4.2`\n- String, e.g. `\"Hello, world\"`\n- Bitstring (binary), e.g. ` >` (encodes to an OSC Blob)\n- Atom, e.g. `:tremelo` (encodes to an OSC Symbol)\n- `%{midi: value}` (encodes the the *4-byte* value to an OSC MIDI type)\n- `%{seconds: seconds, fraction: fraction}` or `%{time: value}` where `value` is a 64-bit integer (encodes to an OSC Time tag)\n- `%{rgba: [r, g, b, a]}` (encodes to an RGBA colour where `r`, `g`, `b` and `a` are all integers)\n- List, e.g. `[1, 2, 3]` (encodes to an OSC Array)\n- `true` which encodes to the OSC type: True\n- `false` which encodes to the OSC type: False\n- `nil` or `:null` which encodes to the OSC type: Null\n- `:impulse` which encodes to the OSC type: Impulse\n\nThe two main functions are:\n- `encode/1` which takes an `%OSCx.Message{}` struct and encodes it to the OSC message format\n- `decode/1` which takes an OSC message recieved (e.g. via UDP) and decodes it into an `%OSCx.Message{}` struct.\n\nTo learn more about how this library encodes and decodes OSC data, see: [Arguments and types](arguments_and_types.md).","ref":"OSCx.Message.html"},{"type":"module","title":"Structure of OSC messages - OSCx.Message","doc":"OSC messages are made up of three parts:\n1. **Address** which represents the function you want to control starting with a forward slash e.g. `\"/status\"`\n2. **Tag type string**: which lists the data types in the data payload, in the order they occur. Note that some older implementations of OSC may omit the OSC Type Tag string.\n3. **Arguments**: the data payload which could be any of the OSC types.\n\n![OSC message diagram](assets/osc-message.png)","ref":"OSCx.Message.html#module-structure-of-osc-messages"},{"type":"module","title":"Creating a message - OSCx.Message","doc":"When you create a message, you only need to specify the address and the arguments like this:\n```\niex> my_msg = %OSCx.Message{address: \"/target/address\", arguments: [1, 2.0, \"string data\"]}\n```\nThe tag type string is automatically generated when the message is encoded:\n```\n# Encode the message above\niex> encoded_msg = my_msg |> OSCx.encode()\n\n# Inspect the message to see how it is encoded\n# The OSCx.Decoder.inspect() shows:\n# - the character code of each byte\n# - its printable utf8 value\n# - underscore (_) is used to denote either a 0 value or padding\n# - (D) is a non-printable data value\niex> encoded_msg |> OSCx.Decoder.inspect()\n[\n  {0, [\"47 '/'\", \"116 't'\", \"97 'a'\", \"114 'r'\"]},\n  {1, [\"103 'g'\", \"101 'e'\", \"116 't'\", \"47 '/'\"]},\n  {2, [\"97 'a'\", \"100 'd'\", \"100 'd'\", \"114 'r'\"]},\n  {3, [\"101 'e'\", \"115 's'\", \"115 's'\", \"0 (_)\"]},\n  {4, [\"44 ','\", \"105 'i'\", \"102 'f'\", \"115 's'\"]},\n  {5, [\"0 (_)\", \"0 (_)\", \"0 (_)\", \"0 (_)\"]},\n  {6, [\"0 (_)\", \"0 (_)\", \"0 (_)\", \"1 (D)\"]},\n  {7, [\"64 '@'\", \"0 (_)\", \"0 (_)\", \"0 (_)\"]},\n  {8, [\"115 's'\", \"116 't'\", \"114 'r'\", \"105 'i'\"]},\n  {9, [\"110 'n'\", \"103 'g'\", \"32 ' '\", \"100 'd'\"]},\n  {10, [\"97 'a'\", \"116 't'\", \"97 'a'\", \"0 (_)\"]}\n]\n```\nYou can see on row number 4, the automatically encoded type string of `,ifs` which means integer, float and string which was created for the OSC message arguments of `[1, 2.0, \"string data\"]`.","ref":"OSCx.Message.html#module-creating-a-message"},{"type":"module","title":"Transmitting and receiving messages - OSCx.Message","doc":"OSC messages are independent from any specific transport mechanism, and can be transmitted and receivied over a variety of networks, including Ethernet, Wi-Fi, and the Internet.\n\n#","ref":"OSCx.Message.html#module-transmitting-and-receiving-messages"},{"type":"module","title":"Using UDP - OSCx.Message","doc":"Even though OSC messages are transport mechanism agnostic, they are commonly sent and received using UDP sockets.\n\nThe Erlang `:gen_utp` module can be used for this purpose. It provides the functions necessary for communicating with sockets using the UDP protocol.","ref":"OSCx.Message.html#module-using-udp"},{"type":"module","title":"Example - OSCx.Message","doc":"```\n# IP or host and port number for the UDP connection\nip_address = '127.0.0.1' # This could be changed to named address, like 'localhost'\nport_num = 57110 # In this example, this is the default port used by SuperCollider\n\n# Open a port\n{:ok, port} = :gen_udp.open(0, [:binary, {:active, true}])\n\n# Encode the message\nosc_message = %OSCx.Message{address: \"/version\", arguments: []} |> OSCx.encode()\n\n# Send message\n:gen_udp.send(port, ip_address, port_num, osc_message)\n```","ref":"OSCx.Message.html#module-example"},{"type":"module","title":"More information - OSCx.Message","doc":"See the OSC specification website at: https://opensoundcontrol.stanford.edu/index.html","ref":"OSCx.Message.html#module-more-information"},{"type":"function","title":"OSCx.Message.decode/1","doc":"Decodes a binary OSC message.","ref":"OSCx.Message.html#decode/1"},{"type":"function","title":"Example - OSCx.Message.decode/1","doc":"```\n# Binary OSC message\niex> bin_msg =  >\n >\n\niex> decoded_msg = OSCx.Message.decode(bin_msg)\n%OSCx.Message{address: \"/status\", arguments: []}\n```\nIn practice, use `OSCx.decode/1` instead which can accept messages or bundles.","ref":"OSCx.Message.html#decode/1-example"},{"type":"function","title":"OSCx.Message.encode/1","doc":"Encodes an `%OSCx.Message{}` struct to an OSC binary message.\n\nThis function takes a populated `%OSCx.Message{}` struct as its first and only parameter.\n\nIt returns a binary message in the format:\n\n> ** ** followed by an ** ** followed by ** **.\n\nIn practice, use `OSCx.encode/1` instead which can accept messages or bundles.","ref":"OSCx.Message.html#encode/1"},{"type":"function","title":"Example - OSCx.Message.encode/1","doc":"```\niex> %OSCx.Message{address: \"/status\", arguments: []} |> OSCx.Message.encode()\n >\n```","ref":"OSCx.Message.html#encode/1-example"},{"type":"function","title":"OSCx.Message.new/1","doc":"Convenience function for creating an `%OSCx.Message{}` struct.\n\nOptionally takes a keyword list as it's first parameter, recognising any of the following keys:\n- `address:` the OSC adress string, for example `address: \"/synth/play_note\"`\n- `arguments:` a list of arguments, for example `arguments: [440.5]`","ref":"OSCx.Message.html#new/1"},{"type":"function","title":"Example - OSCx.Message.new/1","doc":"```\niex> OSCx.Message.new(address: \"/synth/play_chord\", arguments: [440.0, 445.0, 450.0])\n\n%OSCx.Message{\n  address: \"/synth/play_chord\",\n  arguments: [440.0, 445.0, 450.0],\n}\n```","ref":"OSCx.Message.html#new/1-example"},{"type":"module","title":"OSCx.Decoder","doc":"Helpers to decode an OSC message into Elixir.\n\nThis module contains helper functions that parse OSC binary data and converts them into equivalent Elixir types.\n\nTo decode the full OSC data, these functions are chained together and are used by `OSCx.Message.decode/1` and `OSCx.Bundle.decode/1`.\n\n> ##","ref":"OSCx.Decoder.html"},{"type":"module","title":"Tip {: .tip} - OSCx.Decoder","doc":">\n> Note that in practice you'll will likely not need to use this module directly and instead use `OSCx.decode/1`.","ref":"OSCx.Decoder.html#module-tip-tip"},{"type":"module","title":"Automatically decoded types - OSCx.Decoder","doc":"#","ref":"OSCx.Decoder.html#module-automatically-decoded-types"},{"type":"module","title":"Types with arguments - OSCx.Decoder","doc":"The following types can be automatically decoded:\n\n| Type tag | Unicode number | Type            | OSC spec version |\n| -------- | -------------- | --------------- | ---------------- |\n| i        | 105            | 32 bit integer  | 1.0+ required    |\n| f        | 102            | 32 bit float    | 1.0+ required    |\n| s        | 115            | String          | 1.0+ required    |\n| b        | 98             | Blob            | 1.0+ required    |\n| h        | 104            | 64-bit big-endian two’s complement integer | 1.0 non-standard |\n| d        | 100            | 64 bit (“double”) IEEE 754 floating point number | 1.0 non-standard |\n| m        | 109            | 4 byte MIDI message | 1.0 non-standard |\n| t        | 116            | OSC time tag    | 1.1+ required    |\n| r        | 144            | 32 bit RGBA color | 1.0 non-standard |\n| c        | 99             | An ascii character, sent as 32 bits | 1.0 non-standard |\n| S        | 83             | Symbol          | 1.0 non-standard    |\n| [ and ]  | 91 and 93      | List            | 1.0 non-standard    |\n\n#","ref":"OSCx.Decoder.html#module-types-with-arguments"},{"type":"module","title":"Symbols, Atoms and Strings - OSCx.Decoder","doc":"By default, Symbols are converted to Elixir atoms. This behaviour can be changed, see `OSCx.Decoder.set_symbol_to_atom/1`.\n\n#","ref":"OSCx.Decoder.html#module-symbols-atoms-and-strings"},{"type":"module","title":"Special string tag types - OSCx.Decoder","doc":"The following string type tags are also automatically decoded via the `tag/1` function:\n\n| Type tag | Unicode number | Type            | OSC spec version |\n| -------- | -------------- | --------------- | ---------------- |\n| T        | 84             | True (tag only, no arguments) | 1.1+ required |\n| F        | 80             | False (tag only, no arguments) | 1.1+ required |\n| N        | 78             | Null (tag only, no arguments) | 1.1+ required |\n| I        | 73             | Impulse, also known as Infinitum in the OSC 1.0 Spec, or 'Bang' (tag only, no arguments) | 1.1+ required |","ref":"OSCx.Decoder.html#module-special-string-tag-types"},{"type":"module","title":"More information - OSCx.Decoder","doc":"To learn more about how this library decodes and encodes data, see: [Arguments and types](arguments_and_types.md).","ref":"OSCx.Decoder.html#module-more-information"},{"type":"function","title":"OSCx.Decoder.address/1","doc":"Extracts the address from the binary.\n\nThis function assumes the OSC binary is in the following format:\n\n>   +   +  \n\nIt returns a tuple with the first element containing the OSC address, and the second the remaining OSC data which will be the tag type string followed by OSC arguments.","ref":"OSCx.Decoder.html#address/1"},{"type":"function","title":"OSCx.Decoder.blob/2","doc":"Extracts a blob (non-string binary sequence) from the binary.\n\nThis function assumes that the binary starts with the blob type.\n\nAny additional padding added to the blob is removed.\n\nReturns a tuple with the:\n- first element: the blob value\n- second element: the remaining binary data\n- third element: the rest of the type tags to be processed.\n\nThis is similar to the `OSCx.Decoder.string/2` function.","ref":"OSCx.Decoder.html#blob/2"},{"type":"function","title":"OSCx.Decoder.char/2","doc":"Decodes a 32-bit char value.\n\nReturns it as an Elixir charlist, e.g.:application\n```\n# This is the char 'a' in 32-bit\niex> binary =  >\n >\n\niex> {char, _rem_bin, _rem_tags} = OSCx.Decoder.char(binary, [])\n{~c\"a\", \"\", []}\n\niex> char\n~c\"a\"\n```","ref":"OSCx.Decoder.html#char/2"},{"type":"function","title":"OSCx.Decoder.de_pad/2","doc":"Removes padding prepended to remaining binary data.\n\nOSC messages are encoded in chunks of 4 bytes. When arbitary data, such as a string or blob is less that 4 bytes, padding using a value of ` >` is added.\n\nThis function takes the following parameters:\n- String or Blob (both are binary data of an arbitary length) OR byte size of the string or blob\n- Remaning binary data once the String or Blob has been extracted\n\nReturns the remaing binary data with any leading padding removed.","ref":"OSCx.Decoder.html#de_pad/2"},{"type":"function","title":"Example - OSCx.Decoder.de_pad/2","doc":"#","ref":"OSCx.Decoder.html#de_pad/2-example"},{"type":"function","title":"String - OSCx.Decoder.de_pad/2","doc":"```\niex> bin_string =  >\n >\n\niex> [string, rest] = :binary.split(bin_string,  >)\n[\"Hello\",  >]\n\n# Leading padding will be removed, based on the previous strings size\niex> OSCx.Decoder.de_pad(string, rest)\n >\n```\n#","ref":"OSCx.Decoder.html#de_pad/2-string"},{"type":"function","title":"Byte size - OSCx.Decoder.de_pad/2","doc":"```\niex> bin_string =  >\n >\n\niex> [string, rest] = :binary.split(bin_string,  >)\n[\"Hello\",  >]\n\n# Set byte size to 5 (\"Hello\" has byte size of 5). Any leading padding will be removed based on this size.\niex> OSCx.Decoder.de_pad(5, rest)\n >\n```","ref":"OSCx.Decoder.html#de_pad/2-byte-size"},{"type":"function","title":"OSCx.Decoder.decode_arg/3","doc":"Decodes arguments from the binary.\n\nCurrently this automatically decodes the following argument types:\n\n| Type tag | Unicode number | Type            | OSC spec version |\n| -------- | -------------- | --------------- | ---------------- |\n| i        | 105            | 32 bit integer  | 1.0+ required    |\n| f        | 102            | 32 bit float    | 1.0+ required    |\n| s        | 115            | String          | 1.0+ required    |\n| b        | 98             | Blob            | 1.0+ required    |\n| h        | 104            | 64-bit big-endian two’s complement integer | 1.0 non-standard |\n| d        | 100            | 64 bit (“double”) IEEE 754 floating point number | 1.0 non-standard |\n| m        | 109            | 4 byte MIDI message | 1.0 non-standard |\n| c        | 99             | An ascii character, sent as 32 bits | 1.0 non-standard |\n| t        | 116            | OSC time tag    | 1.1+ required    |\n| r        | 144            | 32 bit RGBA color | 1.0 non-standard |\n| S        | 83             | Symbol          | 1.0 non-standard    |\n| [ and ]  | 91 and 93      | List            | 1.0 non-standard    |\n\nThis function takes the following parameters:\n- List (charlist) of tag types (e.g. ?s for string, ?i for integer, etc)\n- Binary data, with the address and tag type string removed so that only argument data remains\n\nReturns a tuple with the first element containing the decoded data, and the second element any remaining data. The second element will be an empty binary if all data was successfully decoded to the end.","ref":"OSCx.Decoder.html#decode_arg/3"},{"type":"function","title":"OSCx.Decoder.float/2","doc":"Decodes a 32-bit float from the head of the binary.\n\nReturns a tuple with the first element the float value, and the second element the remaining binary data.\n\nReturns a tuple with the:\n- first element: the float value\n- second element: the remaining binary data\n- third element: the rest of the type tags to be processed.","ref":"OSCx.Decoder.html#float/2"},{"type":"function","title":"OSCx.Decoder.float_64bit/2","doc":"Decodes a 64-bit float from the head of the binary.\n\nReturns a tuple with the first element the float value, and the second element the remaining binary data.\n\nReturns a tuple with the:\n- first element: the float value\n- second element: the remaining binary data\n- third element: the rest of the type tags to be processed.","ref":"OSCx.Decoder.html#float_64bit/2"},{"type":"function","title":"OSCx.Decoder.inspect/1","doc":"Generates a list showing each 'row' of the encoded OSC message.\n\nThis maybe useful in debugging OSC binary messages.\n\nThe OSCx.Decoder.inspect() shows:\n- the character code of each byte\n- its printable utf8 value\n- underscore (_) is used to denote either a 0 value or padding\n- (D) is a non-printable data value","ref":"OSCx.Decoder.html#inspect/1"},{"type":"function","title":"Example - OSCx.Decoder.inspect/1","doc":"```\niex> encoded_msg =\n  %OSCx.Message{address: \"/target/address\", arguments: [1, 2.0, \"string data\"]}\n  |> OSCx.encode()\n\n# Inspect the message to see how it is encoded\niex> encoded_msg |> OSCx.Decoder.inspect()\n[\n  {0, [\"47 '/'\", \"116 't'\", \"97 'a'\", \"114 'r'\"]},\n  {1, [\"103 'g'\", \"101 'e'\", \"116 't'\", \"47 '/'\"]},\n  {2, [\"97 'a'\", \"100 'd'\", \"100 'd'\", \"114 'r'\"]},\n  {3, [\"101 'e'\", \"115 's'\", \"115 's'\", \"0 (_)\"]},\n  {4, [\"44 ','\", \"105 'i'\", \"102 'f'\", \"115 's'\"]},\n  {5, [\"0 (_)\", \"0 (_)\", \"0 (_)\", \"0 (_)\"]},\n  {6, [\"0 (_)\", \"0 (_)\", \"0 (_)\", \"1 (D)\"]},\n  {7, [\"64 '@'\", \"0 (_)\", \"0 (_)\", \"0 (_)\"]},\n  {8, [\"115 's'\", \"116 't'\", \"114 'r'\", \"105 'i'\"]},\n  {9, [\"110 'n'\", \"103 'g'\", \"32 ' '\", \"100 'd'\"]},\n  {10, [\"97 'a'\", \"116 't'\", \"97 'a'\", \"0 (_)\"]}\n]\n```","ref":"OSCx.Decoder.html#inspect/1-example"},{"type":"function","title":"OSCx.Decoder.integer/2","doc":"Decodes a 32-bit integer from the head of the binary.\n\nReturns a tuple with the:\n- first element: the integer value\n- second element: the remaining binary data\n- third element: the rest of the type tags to be processed.","ref":"OSCx.Decoder.html#integer/2"},{"type":"function","title":"OSCx.Decoder.integer_64bit/2","doc":"Decodes a 64-bit integer from the head of the binary.\n\nReturns a tuple with the:\n- first element: the integer value\n- second element: the remaining binary data\n- third element: the rest of the type tags to be processed.","ref":"OSCx.Decoder.html#integer_64bit/2"},{"type":"function","title":"OSCx.Decoder.list/2","doc":"Parses a list (or array as it is called in OSC).\n\nReturns a tuple with the:\n- first element: a list containing the decoded array data\n- second element: the remaining binary data\n- third element: the rest of the type tags to be processed.","ref":"OSCx.Decoder.html#list/2"},{"type":"function","title":"OSCx.Decoder.midi/2","doc":"Extracts a MIDI message from the head of the binary.\n\nThis function assumes that the binary starts with a MIDI message.\n\nReturns a tuple with the:\n- first element: a map with a 4-byte MIDI value, in the format: `%{midi: value}`\n- second element: the remaining binary data\n- third element: the rest of the type tags to be processed.","ref":"OSCx.Decoder.html#midi/2"},{"type":"function","title":"OSCx.Decoder.rgba/2","doc":"Extracts a MIDI message from the head of the binary.\n\nThis function assumes that the binary starts with a MIDI message.\n\nReturns a tuple with the:\n- first element: a map with a 4-byte MIDI value, in the format: `%{midi: value}`\n- second element: the remaining binary data\n- third element: the rest of the type tags to be processed.","ref":"OSCx.Decoder.html#rgba/2"},{"type":"function","title":"OSCx.Decoder.set_symbol_to_atom/1","doc":"Sets the default behaviour when decoding of OSC Symbol types.\n\nTakes as its first parameter a boolean representing the following behaviour:\n- true: OSC Symbols are converted to Elixir atoms. This is the default behaviour.\n- false: OSC Symbols are converted to Elixir strings.\n\nThis function changes the `:OSCx, :symbol_to_atom` application evironment, e.g.: `Application.put_env(:OSCx, :symbol_to_atom, false)`.","ref":"OSCx.Decoder.html#set_symbol_to_atom/1"},{"type":"function","title":"Why this option? - OSCx.Decoder.set_symbol_to_atom/1","doc":"Uncontrolled dynamic creation of atoms on the BEAM may not be considered good practice.\nAtoms are not garbage collected by the Erlang Virtual Machine, so they live in memory during a software's entire execution lifetime.\nAlthough there is a high-limit on the number of atoms that can be created on the BEAM, for situations where there may be a large variety of atoms, converting to Elixir's string type may be preferrable.\nFor more information on this see [Code-related anti-patterns](https://hexdocs.pm/elixir/main/code-anti-patterns.html).","ref":"OSCx.Decoder.html#set_symbol_to_atom/1-why-this-option"},{"type":"function","title":"OSCx.Decoder.special_tag/3","doc":"Decodes 'special' string tag types without arguments, such as: True, False, Null and Impulse.\n\nThis function is used by `decode_arg/3` as part of the decoding process. It takes the following parameters:\n- argument data (binary)\n- charlist of remaing tags to process\n- the value to return (e.g. `true` for `?T`)\n\nOSC specifies a number of tags which don't carry arguments. These are:\n\n| Type tag | Unicode number | OSC Type                      | OSC spec version | Elixir value |\n| -------- | -------------- | ----------------------------- | ---------------- | ------------ |\n| T        | 84             | True (tag only, no arguments) | 1.1+ required    | `true`       |\n| F        | 80             | False (tag only, no arguments)  | 1.1+ required  | `false`      |\n| N        | 78             | Null (tag only, no arguments) | 1.1+ required    | `nil`        |\n| I        | 73             | Impulse, also known as Infinitum in the OSC 1.0 Spec, or 'Bang' (tag only, no arguments) | 1.1+ required | `:impulse` |\n\nReturns a tuple with the:\n- first element: the decoded value as an Elixir type\n- second element: the remaining binary data (unlike other functions, this leaves the binary data unchanged)\n- third element: the rest of the type tags to be processed (this is also left unchanged).","ref":"OSCx.Decoder.html#special_tag/3"},{"type":"function","title":"OSCx.Decoder.string/2","doc":"Extracts a string from the binary.\n\nThis function assumes that the binary starts with a string type.\n\nAny additional padding added to the string is removed.\n\nReturns a tuple with the:\n- first element: the string value\n- second element: the remaining binary data\n- third element: the rest of the type tags to be processed.\n\nThis is similar to the `OSCx.Decoder.blob/2` function.","ref":"OSCx.Decoder.html#string/2"},{"type":"function","title":"Example - OSCx.Decoder.string/2","doc":"```\niex> bin_string =  >\n >\n\niex> OSCx.Decoder.string(bin_string)\n{\"Hello, world!\", \"\"}\n```","ref":"OSCx.Decoder.html#string/2-example"},{"type":"function","title":"OSCx.Decoder.symbol/2","doc":"Extracts a symbols from the binary.\n\nIn the OSC 1.0 Specification, this is considered an alternate OSC-string representation, used for example in systems that differentiate “symbols” from “strings”.\n\n##\nThis function assumes that the binary starts with a symbol type.\n\nAny additional padding added to the symbol is removed.\n\nReturns a tuple with the:\n- first element: the symbol as an atom (or string if default behaviour is changed, see below)\n- second element: the remaining binary data\n- third element: the rest of the type tags to be processed.\n\nThis is similar to the `OSCx.Decoder.string/2` function.","ref":"OSCx.Decoder.html#symbol/2"},{"type":"function","title":"Symbols become `:atoms` by default - OSCx.Decoder.symbol/2","doc":"In this library, Elixir's atom type is considered an OSC Symbol type.\n\nUncontrolled creation of atoms is not considered good pratice on the BEAM. If there are potentially a lot of Symbols being decoded, consider decoding to a String instead.\n\nThis default behaviour can be changed by calling `OSCx.Decoder.set_symbol_to_atom(false)` or setting the application environment `Application.get_env(:OSCx, :symbol_to_atom, false)`.\n\nFor more information, see `OSCx.Decoder.set_symbol_to_atom/1`.","ref":"OSCx.Decoder.html#symbol/2-symbols-become-atoms-by-default"},{"type":"function","title":"OSCx.Decoder.tag/1","doc":"Parses a tag string for special tag types.\n\nTakes a type tag string as the first parameter, with or without the leading comma.\n\nParses a tag string to check if any of the following special types are present:\n\n| OSC string tag type | OSC character | Elixir type |\n| True | T | true |\n| False | F | false |\n| Null | N | nil |\n| Impulse or Infinitum or 'Bang' | I | :impulse |\n\nReturn type is a list of zero or more of the above.","ref":"OSCx.Decoder.html#tag/1"},{"type":"function","title":"Example - OSCx.Decoder.tag/1","doc":"```\n# Single 'true' type present\niex> Decoder.tag(\",T\")\n[true]\n\n# Single 'false' type present\niex> Decoder.tag(\",F\")\n[false]\n\n# Single 'null' type present, will return nil in Elixir\niex> Decoder.tag(\",N\")\n[nil]\n\n# Multiple special types given.\niex> Decoder.tag(\",FTNI\")\n[false, true, nil, :impulse]\n\n # No special types present, other tags ignored. Empty list is returned.\niex> Decoder.tag(\",ifs\")\n[]\n```","ref":"OSCx.Decoder.html#tag/1-example"},{"type":"function","title":"OSCx.Decoder.tag_list/1","doc":"Extracts the list of tags.\n\nTakes the binary data from the message. This assumes that the address has been removed from the start of the OSC message, and only the tag type string and arguments are remaining:\n\n>   +   +  \n\nReturns a tuple with the first element containing the tags (with the leading comma removed), and the second element the remaining binary data with tag type string and it's padding removed.","ref":"OSCx.Decoder.html#tag_list/1"},{"type":"function","title":"OSCx.Decoder.time/2","doc":"Extracts a time-tag from the head of the binary.\n\nThis function assumes that the binary starts with a time-tag.\n\nReturns a tuple with the:\n- first element: the map value for a time tag (see below)\n- second element: the remaining binary data\n- third element: the rest of the type tags to be processed.\n\nThe time tag map is in the format:\n```\n%{time: value, seconds: seconds, fraction: fraction}\n```\nWhere:\n- `time` is the full 64-bit integer time code\n- `seconds` is the number of seconds since midnight on January 1, 1900 (32-bit integer)\n- `fraction` is the fractional part of a second to a precision of about 200 picoseconds (32-bit integer).\n\nThe 32-bit `:seconds` and `:fraction` have been extracted from the 64-bit `:time` value, so you can pattern match on whatever is most useful in the context of your application.","ref":"OSCx.Decoder.html#time/2"},{"type":"module","title":"OSCx.Encoder","doc":"Helpers to encode Elixir types into OSC data types.\n\nThis module contains helper functions used to encode supported Elixir types (see below) into OSC binary data.\n\nTo encode the full OSC data, these functions are chained together and are used by `OSCx.Message.encode/1` and `OSCx.Bundle.encode/1`.\n\n> ##","ref":"OSCx.Encoder.html"},{"type":"module","title":"Tip {: .tip} - OSCx.Encoder","doc":">\n> Note that in practice you'll will likely not need to use this module directly and instead use `OSCx.encode/1`.\n\nAt it's core, OSC has the following data types:\n\n| Type        | Description |\n| ----------- | ----------- |\n| **Integer** | a 32-bit signed integer and also a 64-bit integer type) |\n| **Float**   | a 32-bit IEEE 754 floating-point number and also a 64-bit double type |\n| **String**  | a sequence of printable ASCII characters (there are also Symbols and Char types)|\n| **Blob**    | a sequence of arbitrary binary data, with its size prepended |\n| **Timetag** | a 64-bit fixed-point number that represents a time in seconds since midnight on January 1, 1900. The first 32 bits of the timetag represent the number of seconds, and the last 32 bits represent fractional parts of a second to a precision of about 200 picoseconds. |\n\nThe above are considered 'required' types by the OSC spec, however, OSC can be extended to support additional or optional types.\n\nThe main function in this module is `encode_arg/1`, which takes one of the recognised Elixir data types, and returns it as an OSC type.","ref":"OSCx.Encoder.html#module-tip-tip"},{"type":"module","title":"Tag-specific types - OSCx.Encoder","doc":"Additionally the [OSC 1.1 Specification](https://opensoundcontrol.stanford.edu/spec-1_1.html) includes some types which carry no data arguments, but can be encoded directly into the tag type string. These are True, False, Null and Impluse.\n\nSee the [Tag specific](#tag-specific) functions below.","ref":"OSCx.Encoder.html#module-tag-specific-types"},{"type":"module","title":"4-byte boundaries - OSCx.Encoder","doc":"OSC data is aligned on 4-byte boundaries. 32-bit types like Integers and Floats are 4-bytes, but Strings and Blobs can be of an arbintary length. For this reason they may be padded with extra null characters (` >>` is used in Elixir) to make the total length a multiple of 4 bytes.","ref":"OSCx.Encoder.html#module-4-byte-boundaries"},{"type":"module","title":"Values returned - OSCx.Encoder","doc":"[Most functions](#type-functions) in this module return a tuple in the form `{type, value}`. These are defined as follows:\n- **type**: first element is an OSC type tag, which is a unicode charater representing the type\n- **value**: second element is the encoded OSC value.\n\nFor example, `{105,  >}` is returned for encoding a the integer `128`. See the example below.","ref":"OSCx.Encoder.html#module-values-returned"},{"type":"module","title":"Type tags - OSCx.Encoder","doc":"OSC type tags are unicode characters which represent each type. This library currently implements the following types:\n\n| Type tag | Unicode number | Type            | OSC spec version |\n| -------- | -------------- | --------------- | ---------------- |\n| i        | 105            | 32 bit integer  | 1.0+ required    |\n| f        | 102            | 32 bit float    | 1.0+ required    |\n| s        | 115            | String          | 1.0+ required    |\n| b        | 98             | Blob            | 1.0+ required    |\n| h        | 104            | 64-bit big-endian two’s complement integer | 1.0 non-standard |\n| d        | 100            | 64 bit (“double”) IEEE 754 floating point number | 1.0 non-standard |\n| c        | 99             | An ascii character, sent as 32 bits | 1.0 non-standard |\n| m        | 109            | 4 byte MIDI message | 1.0 non-standard |\n| t        | 116            | OSC time tag    | 1.1+ required    |\n| r        | 144            | 32 bit RGBA color | 1.0 non-standard |\n| [ and ]  | 91 and 93      | List            | 1.0 non-standard    |\n| T        | 84             | True (tag only, no arguments) | 1.1+ required |\n| F        | 80             | False (tag only, no arguments) | 1.1+ required |\n| N        | 78             | Null (tag only, no arguments) | 1.1+ required |\n| I        | 73             | Impulse (tag only, no arguments) | 1.1+ required |\n\nThese type tags are used to build a type tag string, which is part of the OSC message.","ref":"OSCx.Encoder.html#module-type-tags"},{"type":"module","title":"Example - OSCx.Encoder","doc":"```\niex> alias OSCx.Encoder\n\niex> my_integer = 128\niex> Encoder.integer(my_integer)\n# Returns the type 105 (integer), and the encoded value:\n{105,  >}\n\niex> very_big_integer = 9_223_372_036_854_775_700\niex> Encoder.integer(very_big_integer)\n# Returns the type 104 (64-bit big-endian two’s complement integer), with the encoded value\n{104,  >}\n\niex> my_string = \"hello world\"\niex> Encoder.string(my_string)\n# Returns the type 115 (string), and a list containing the string with any required padding:\n{115,  >}\n```","ref":"OSCx.Encoder.html#module-example"},{"type":"module","title":"More information - OSCx.Encoder","doc":"To learn more about how this library encodes and decodes data, see: [Arguments and types](arguments_and_types.md).","ref":"OSCx.Encoder.html#module-more-information"},{"type":"function","title":"OSCx.Encoder.blob/1","doc":"OSC-blob: a 32-bit size count followed by that many bytes of arbitrary binary data (total must be a multiple of 4) – flexibility to send any encoding","ref":"OSCx.Encoder.html#blob/1"},{"type":"function","title":"OSCx.Encoder.char/1","doc":"Char: Encode a character as 32 bit.\n\nIn OSC 1.0, the Char type is optional and is used to encode an ascii character in 32 bits.\n\nThis function accepts a map `%{char: value}` as the first parameter, where value can be a\n- single character string, e.g. `\"A\"`\n- single charlist, e.g. `~c\"A\"` or `'A'`\n- an integer representing an ASCII char, e.g. `65`.","ref":"OSCx.Encoder.html#char/1"},{"type":"function","title":"Example - OSCx.Encoder.char/1","doc":"All of the following are equivalent for the ASCII 'A' character:\n```\niex> %{char: \"A\"} |> OSCx.Encoder.char()\n{99,  >}\n\niex-> %{char: ~c\"A\"} |> OSCx.Encoder.char()\n{99,  >}\n\niex-> %{char: 'A'} |> OSCx.Encoder.char()\n{99,  >}\n\niex-> %{char: 65} |> OSCx.Encoder.char()\n{99,  >}\n```","ref":"OSCx.Encoder.html#char/1-example"},{"type":"function","title":"OSCx.Encoder.encode_arg/1","doc":"Encodes an argument into its OSC type and value.","ref":"OSCx.Encoder.html#encode_arg/1"},{"type":"function","title":"Elixir to OSC types - OSCx.Encoder.encode_arg/1","doc":"You can call this function with the following Elixir types:\n\n| Elixir type | OSC type tag | Unicode number | OSC type        |\n| ----------- | ------------ | -------------- | --------------- |\n| Integer     | i            | 105            | 32 bit integer  |\n| Float       | f            | 102            | 32 bit float    |\n| String      | s            | 115            | String          |\n| Atom        | s            | 115            | String          |\n| Bitstring   | b            | 98             | Blob            |\n| Integer (64 bit) | h       | 104            | 64-bit big-endian two’s complement integer |\n| Float (64 bit) | d         | 104            | 64-bit big-endian two’s complement integer |\n| Time map    | t            | 116            | OSC time tag    |\n| MIDI map    | m            | 109            | 4-byte MIDI message |\n| Char map    | c            | 99             | 4-byte ASCII character |\n\nNote the:\n- time map is in the format `%{seconds: seconds, fraction: fraction}` where seconds and fraction are 32-bit numbers.\n- MIDI map is in the format `%{midi: value}` where the value is a 4 byte MIDI message\n- Char map is in the format `%{char: value}` where value is a single character string (e.g. `\"A\"`), a single char (e.g. `'A'` or `~c\"A\"`) or an integer of an ASCII char (e.g. `65`).","ref":"OSCx.Encoder.html#encode_arg/1-elixir-to-osc-types"},{"type":"function","title":"Return format - OSCx.Encoder.encode_arg/1","doc":"The function will return a tuple with the first element containing the OSC type tag, and the second element the OSC encoded value such as `{105,  >}`.\n\nIf an elixir List is given, each argument in the list will be encoded and a list is returned, for example:\n```\niex> Encoder.encode_arg([1,2,3])\n[{105,  >}, {105,  >}, {105,  >}]\n```\n\nAn unrecognised Elixir type will return an error tuple in the following format:\n```\n{:error, \"Unknown type\"}\n```","ref":"OSCx.Encoder.html#encode_arg/1-return-format"},{"type":"function","title":"Examples - OSCx.Encoder.encode_arg/1","doc":"```\nalias OSCx.Encoder\n\n# Encode an integer of value 1\niex> Encoder.encode_arg(1)\n{105,  >}\n\n# Encode a float of 2.1\niex> Encoder.encode_arg(2.1)\n{102,  >}\n\n# Encode a string of \"hello world\" (this will also add padding do the end)\n# Bit\niex> Encoder.encode_arg(\"hello world\")\n{115,  >}\n\n# Encode a 64-bit integer\niex> Encoder.encode_arg(9_223_372_036_854_775_700)\n{104,  >}\n```","ref":"OSCx.Encoder.html#encode_arg/1-examples"},{"type":"function","title":"OSCx.Encoder.encoded_value/1","doc":"Returns the encoded values from the encoded argument list.\n\ne.g. returns the `value` from the `{type, value}` tuple as a list.","ref":"OSCx.Encoder.html#encoded_value/1"},{"type":"function","title":"OSCx.Encoder.float/1","doc":"Float: 32-bit big-endian IEEE 754 floating point number, or 64 bit (“double”) IEEE 754 floating point number.\n\nThis function encodes Elixir's float types as follows:\n\n| Type           | Lower value | Upper value |\n| -------------- | ----------- | ----------- |\n| 32 bit float | 1.175494351e-38 | 3.402823466e+38 |\n| 64 bit float | 2.2250738585072014e-308 | 1.7976931348623158e+308 |\n\nNote the ranges in this table are approximate.","ref":"OSCx.Encoder.html#float/1"},{"type":"function","title":"OSCx.Encoder.integer/1","doc":"Integer: 32-bit two’s complement big-endian or 64 bit big-endian two’s complement integer\n\nThis function encodes Elixir's integer types as follows:\n\n| Type           | Lower value | Upper value | Comment |\n| -------------- | ----------- | ----------- | ------- |\n| 32 bit integer | -2_147_483_647 | 2_147_483_647 | This is 2^31 rounded |\n| 64 bit integer | -9_223_372_036_854_775_808 | 9_223_372_036_854_775_808 | This is 2^63 rounded |","ref":"OSCx.Encoder.html#integer/1"},{"type":"macro","title":"OSCx.Encoder.is_char_map/1","doc":"Guard to test if a map for chars has been provided.","ref":"OSCx.Encoder.html#is_char_map/1"},{"type":"macro","title":"OSCx.Encoder.is_midi_map/1","doc":"Guard to test if a MIDI map has been provided.\nA MIDI map is in the following format: `%{midi: value}`.\nThe value is a 4-byte binary.","ref":"OSCx.Encoder.html#is_midi_map/1"},{"type":"macro","title":"OSCx.Encoder.is_rgba_map/1","doc":"Guard to test if a map for chars has been provided.","ref":"OSCx.Encoder.html#is_rgba_map/1"},{"type":"macro","title":"OSCx.Encoder.is_time_map/1","doc":"Guard to test if a time map has been provided.\nA time map is in the following format: `%{seconds: seconds, fraction: fraction}`.","ref":"OSCx.Encoder.html#is_time_map/1"},{"type":"function","title":"OSCx.Encoder.list/1","doc":"Encodes a list.\n\nReturns a tuple with the first element containing the OSC array type tags, and the second element containing the binary encoded array data.","ref":"OSCx.Encoder.html#list/1"},{"type":"function","title":"OSCx.Encoder.midi/1","doc":"Encodes a 4-byte MIDI message.\n\nTakes as it's first parmeter a map with the key-value pair of: `%{midi: value}`.","ref":"OSCx.Encoder.html#midi/1"},{"type":"function","title":"Format of the value - OSCx.Encoder.midi/1","doc":"The OSC defines the 4 byte MIDI messages as bytes from MSB to LSB are: port id, status byte, data1, data2.\n\nIf only a 3 byte message is provided (e.g. a status byte followed by two data bytes), a port id of  > is prepended to make 4 bytes.\n\nThe MIDI value can be in any of these forms:\n- Binary, e.g. ` >`\n- List, e.g. `[153, 77, 63]`\n- An Elixir type which can be directly encoded to 4-byte binary value\n\n#","ref":"OSCx.Encoder.html#midi/1-format-of-the-value"},{"type":"function","title":"Key: - OSCx.Encoder.midi/1","doc":"- MSB = Most significant byte\n- LSB = Least singificant byte","ref":"OSCx.Encoder.html#midi/1-key"},{"type":"function","title":"OSCx.Encoder.pad/1","doc":"Adds padding to a value based on it's size.\n\nThis is used for types of variable length so that they're equal to 4-bytes in length, or a mutuple of 4-bytes.","ref":"OSCx.Encoder.html#pad/1"},{"type":"function","title":"OSCx.Encoder.prefix_size/2","doc":"Used for prepending the size of a value.\n\nThis is mainly used for prepending the size of a binary.\n\nThe size is encoded in 32 bits, but this can be overridden when needed.","ref":"OSCx.Encoder.html#prefix_size/2"},{"type":"function","title":"OSCx.Encoder.rgba/1","doc":"RGBA: Encides a 32-bit RGBA color.\n\nTakes as it's first parameter an `%{rgba: [r, g, b, a]}` map.\n\nThe map contains 4 integer values, repesenging R, G, B and A in that order.","ref":"OSCx.Encoder.html#rgba/1"},{"type":"function","title":"Example - OSCx.Encoder.rgba/1","doc":"```\niex> OSCx.Encoder.rgba(%{rgba: [255, 255, 60, 20]})\n{114,  >}\n\n```","ref":"OSCx.Encoder.html#rgba/1-example"},{"type":"function","title":"OSCx.Encoder.string/1","doc":"OSC-string: a sequence of non-null ASCII characters followed by 1-4 null characters – total string bytes must be a multiple of 4","ref":"OSCx.Encoder.html#string/1"},{"type":"function","title":"OSCx.Encoder.symbol/1","doc":"OSC-symbol: an alternative to strings in OSC systems, used for 'symbols' which are conceptually equivalent to Atoms in Elixir.","ref":"OSCx.Encoder.html#symbol/1"},{"type":"function","title":"OSCx.Encoder.tag/1","doc":"A helper function to encode a tag based on an Elixir type or special atom.\n\nThis function accepts the following Elixir types and returns the equivalent OSC string tag type character as below:\n\n| Elixir value | OSC type | Character returned for OSC string tag type |\n| true | True | T |\n| false | False | F |\n| nil | Null | N |\n| :null | Null | N |\n| :impulse | Impulse (also known as Infinitum in OSC 1.0 Spec, or 'Bang') | I |","ref":"OSCx.Encoder.html#tag/1"},{"type":"function","title":"OSCx.Encoder.tag_false/0","doc":"Used to encode *false* into the tag type string.\n\nTrue equates to the charater 'F'.","ref":"OSCx.Encoder.html#tag_false/0"},{"type":"function","title":"OSCx.Encoder.tag_impulse/0","doc":"Used to encode *impulse* into the tag type string.\n\nImpulse is also known as Infinitum in OSC 1.0 Spec, or 'Bang'.\n\nTrue equates to the charater 'I'.","ref":"OSCx.Encoder.html#tag_impulse/0"},{"type":"function","title":"OSCx.Encoder.tag_null/0","doc":"Used to encode *null* into the tag type string.\n\nTrue equates to the charater 'N'.","ref":"OSCx.Encoder.html#tag_null/0"},{"type":"function","title":"OSCx.Encoder.tag_true/0","doc":"Used to encode *true* into the tag type string.\n\nTrue equates to the charater 'T'.","ref":"OSCx.Encoder.html#tag_true/0"},{"type":"function","title":"OSCx.Encoder.time/1","doc":"OSC-timetag: 64 bit, big-endian, fixed-point floating point number\n\nTakes a map in the format of either:\n- `%{seconds: seconds, fraction: fraction}` where `seconds` and `fraction` are 32-bit integers\n- `%{time: value}` where `value` is a 64-bit integer.","ref":"OSCx.Encoder.html#time/1"},{"type":"function","title":"Immediate time - OSCx.Encoder.time/1","doc":"OSC can encode a time tag which means \"process immediately\". This is a time tag with a value of 63 zero bits followed by a one in the least signifigant bit.\n\nYou can set an immediate time by using the `:immediate` atom as follows:\n```\n%{time: :immediate}\n```","ref":"OSCx.Encoder.html#time/1-immediate-time"},{"type":"function","title":"OSC time tag specification - OSCx.Encoder.time/1","doc":"The OSC Specification defines a time tag as a 64-bit fixed-point number that represents a time in seconds since midnight on January 1, 1900.\n\nThe first 32 bits of the timetag represent the number of seconds, and the last 32 bits represent fractional parts of a second to a precision of about 200 picoseconds.\n\nThis is the representation used by Internet NTP timestamps.","ref":"OSCx.Encoder.html#time/1-osc-time-tag-specification"},{"type":"function","title":"OSCx.Encoder.type_tag_string/1","doc":"Converts a list of encoded arguments to an OSC type tag string. Can also take a single encoded argument.","ref":"OSCx.Encoder.html#type_tag_string/1"},{"type":"extras","title":"README","doc":"![Midiex](assets/oscx-elixir-logo.png)\n\n[![Documentation](http://img.shields.io/badge/hex.pm-docs-green.svg?style=flat)](https://hexdocs.pm/oscx)\n[![Package](https://img.shields.io/hexpm/v/oscx.svg)](https://hex.pm/packages/oscx)","ref":"readme.html"},{"type":"extras","title":"About this library - README","doc":"OSCx is an Elixir library for encoding and decoding Open Sound Control (OSC) messages.\n\nOSC is a flexible protocol that can be used for a wide variety of real-time tasks, such as controlling multimedia devices.\n\nThe API of this module is based on the older [osc_ex](https://github.com/camshaft/osc_ex) library. It was designed to be close to a 'drop in' replacement.\n\nNote that this library is deliberately minimilistic and no network transport or process logic is included.","ref":"readme.html#about-this-library"},{"type":"extras","title":"What is Open Sound Control (OSC)? - README","doc":"OSC is a network protocol for real-time communication between computers and other digital devices.\nIt is used to control sound synthesizers, lighting systems, and other multimedia devices.\nOSC is a lightweight and flexible that is well-suited for real-time applications.\nOSC is an open standard, which means that anyone can develop software or hardware that supports OSC.\n\nFor more detailed information see [https://opensoundcontrol.stanford.edu/](https://opensoundcontrol.stanford.edu/index.html).\n\n#","ref":"readme.html#what-is-open-sound-control-osc"},{"type":"extras","title":"Protocol or content format? - README","doc":"Because OSC does not define features typical of *protocols* such as command-response patterns, error handling or negotiation, it may be more accurate to describe OSC as a *content format*, like JSON or XML. An application that uses OSC only needs to be able to parse and encode to and from the OSC format.\n\nNone the less, its versatility makes it useful and why it's incorporated into music performance sofware such as [SuperCollider](https://supercollider.github.io/).","ref":"readme.html#protocol-or-content-format"},{"type":"extras","title":"OSC concepts - README","doc":"Below is a summary of key OSC concepts:\n\n- ##","ref":"readme.html#osc-concepts"},{"type":"extras","title":"Messages - README","doc":"OSC messages are the basic unit of communication in OSC. They consist of an **address**, a **type tag**, and zero or more arguments:\n  - The address is a string that identifies the target of the message, such as `/my_synth/volume` or `/my_controller/button_1`. This is very similar to the resource string or path seen in URLs.\n  - The type tag specifies the type of the arguments, such as integers, floats, strings, or blobs of binary data. For more information see [arguments and types](arguments_and_types.md).\n\n  See `OSCx.Message`.\n\n- ##","ref":"readme.html#messages"},{"type":"extras","title":"Bundles - README","doc":"OSC bundles are a way to send multiple OSC messages in a single packet. This can be useful for improving performance or reducing bandwidth usage.\n\n  See `OSCx.Bundle`.\n\n- ##","ref":"readme.html#bundles"},{"type":"extras","title":"Servers and Clients - README","doc":"OSC servers listen for incoming OSC messages and clients send OSC messages. An application can be both a server and a client at the same time.\n\n  See `OSCx.encode/1` and `OSCx.decode/1` as well as the [OSCx Livebook tour](livebook/oscx_tour.livemd) which demonstrates sending and receiving OSC messages.\n\n- ##","ref":"readme.html#servers-and-clients"},{"type":"extras","title":"Ports - README","doc":"Although OSC is neutral of the transport layer, OSC messages are generally sent and received on UDP ports.\n\n  See the [OSCx Livebook tour](livebook/oscx_tour.livemd) which demonstrates the use of UDP.\n\n- ##","ref":"readme.html#ports"},{"type":"extras","title":"Timetags - README","doc":"OSC timetags can be used to synchronize OSC messages with other events. This is essential for applications such as audio and video production.\n  \n  OSC uses the same format as 64-bit Internet NTP timestamps, where the first 32 bits specify the number of seconds since midnight on January 1, 1900, and the last 32 bits specify fractional parts of a second to a precision of about 200 picoseconds. \n\n  See `OSCx.Encoder.time/1` and [Time tags and synchronisation](time_tags.md).","ref":"readme.html#timetags"},{"type":"extras","title":"Examples - README","doc":"For some simple examples of using this library, see [Examples](examples.md) or the [OSCx Livebook tour](livebook/oscx_tour.livemd).","ref":"readme.html#examples"},{"type":"extras","title":"Installation - README","doc":"#","ref":"readme.html#installation"},{"type":"extras","title":"Adding it to your Elixir project - README","doc":"The package can be installed by adding `oscx` to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:oscx, \"~> 0.1.1\"}\n  ]\nend\n```\n\n#","ref":"readme.html#adding-it-to-your-elixir-project"},{"type":"extras","title":"Using within Livebook and IEx - README","doc":"```elixir\nMix.install([{:oscx, \"~> 0.1.1\"}])\n```\n\nDocumentation can be found at  .","ref":"readme.html#using-within-livebook-and-iex"},{"type":"extras","title":"Examples","doc":"# Examples \n\nThe purpose of this page is to show simple examples using this library.\n\nExamples cover:\n- [Creating a basic message](#creating-a-basic-message)\n- [Creating a basic bundle](#creating-a-basic-bundle)\n- [Working with arguments](#working-with-arguments)\n- [Sending messages](#sending-messages)\n- [Sending bundles](#sending-bundles)\n- [Receiving messages](#receiving-messages)\n\nYou can also explore the [OSCx tour Livebook](livebook/oscx_tour.livemd) for working examples.","ref":"examples.html"},{"type":"extras","title":"Creating a basic message - Examples","doc":"You can create a message by directly populating a `%OSCx.Message{}` struct or by using `OSCx.Message.new/1`:\n\n```\n%OSCx.Message{address: \"/\", arguments: []}\n```\nor\n```\niex> OSCx.Message.new(address: \"/\", arguments: [])\n%OSCx.Message{address: \"/\", arguments: []}\n```\nYou encode this message to an OSC binary format by passing this to the `OSCx.encode/1` function:\n```\niex> %OSCx.Message{address: \"/\", arguments: []} |> OSCx.encode()\n >\n```","ref":"examples.html#creating-a-basic-message"},{"type":"extras","title":"Creating a basic bundle - Examples","doc":"Similarily, you can create a bundle by directly populating a `%OSCx.Bundle{}` struct or by using `OSCx.Bundle.new/1`:\n\n```\n%OSCx.Bundle{\n  time: %{seconds: 80, fraction: 2230576435},\n  elements: [\n    %OSCx.Message{address: \"/one\", arguments: []},\n    %OSCx.Message{address: \"/two\", arguments: []},\n    %OSCx.Message{address: \"/three\", arguments: []},\n  ]\n}\n\n```\nor\n```\niex> OSCx.Bundle.new(\n  time: %{seconds: 80, fraction: 2230576435},\n  elements: [\n    OSCx.Message.new(address: \"/one\", arguments: []),\n    OSCx.Message.new(address: \"/two\", arguments: []),\n    OSCx.Message.new(address: \"/three\", arguments: [])\n  ]\n)\n%OSCx.Bundle{\n  time: %{seconds: 80, fraction: 2230576435},\n  elements: [\n    %OSCx.Message{address: \"/one\", arguments: []},\n    %OSCx.Message{address: \"/two\", arguments: []},\n    %OSCx.Message{address: \"/three\", arguments: []}\n  ]\n}\n```\nYou encode a bundle and its elements by passing it to the `OSCx.encode/1` function:\n```\niex> %OSCx.Bundle{\n  time: %{seconds: 80, fraction: 2230576435},\n  elements: [\n    %OSCx.Message{address: \"/one\", arguments: []},\n    %OSCx.Message{address: \"/two\", arguments: []},\n    %OSCx.Message{address: \"/three\", arguments: []}\n  ]\n} |> OSCx.encode()\n >\n```","ref":"examples.html#creating-a-basic-bundle"},{"type":"extras","title":"Working with arguments - Examples","doc":"Message arguments are added to the arguments list on the `%OSCx.Message{arguments: []}` struct.\n\nBelow is an example with each of the OSC recognised types.\n\n#","ref":"examples.html#working-with-arguments"},{"type":"extras","title":"Example with Integers: - Examples","doc":"```\n%OSCx.Message{address: \"/\", arguments: [1, 2, 3]}\n```\n\n#","ref":"examples.html#example-with-integers"},{"type":"extras","title":"Example with Floats: - Examples","doc":"```\n%OSCx.Message{address: \"/\", arguments: [0.5, 440.5]}\n```\n\n#","ref":"examples.html#example-with-floats"},{"type":"extras","title":"Example with Strings: - Examples","doc":"```\n%OSCx.Message{address: \"/\", arguments: [\"Hello\", \"world\"]}\n```\n\n#","ref":"examples.html#example-with-strings"},{"type":"extras","title":"Example with Bitstrings (binary) or 'blob': - Examples","doc":"```\n%OSCx.Message{address: \"/\", arguments: [ >]}\n```\n\n#","ref":"examples.html#example-with-bitstrings-binary-or-blob"},{"type":"extras","title":"Example with Chars: - Examples","doc":"```\n%OSCx.Message{\n  address: \"/\",\n  arguments: [%{char: \"A\"}, %{char: ~c\"B\"}, %{char: ~c\"C\"}, %{char: 68}]\n}\n```\nChars use the `#{char: value}` map where the value can be a 1 byte String, a Charlist, or the integer value of the char.\n\n#","ref":"examples.html#example-with-chars"},{"type":"extras","title":"Example with Atom or 'Symbol': - Examples","doc":"```\n%OSCx.Message{address: \"/\", arguments: [:loud, :soft]}\n```\n\n#","ref":"examples.html#example-with-atom-or-symbol"},{"type":"extras","title":"Example with MIDI: - Examples","doc":"```\n%OSCx.Message{address: \"/\", arguments: [%{midi: [1, 153, 77, 63]}]}\n```\nThe MIDI format here is `[port_id,  status_byte, data_byte_1, data_byte_2]`. If only a three value message like `%{midi: [153, 77, 63]}` is given, a port_id of 0 is prepended.\n\n#","ref":"examples.html#example-with-midi"},{"type":"extras","title":"Example with RGBA colour: - Examples","doc":"```\n%OSCx.Message{address: \"/\", arguments: [%{rgba: [255, 255, 90, 20]}]}\n```\n\n#","ref":"examples.html#example-with-rgba-colour"},{"type":"extras","title":"Examples with Time tags: - Examples","doc":"Two 32-bit integers representing seconds and fraction of a second:\n```\n%OSCx.Message{\n  address: \"/\",\n  arguments: [%{time: %{seconds: 80, fraction: 2230576435}}]\n}\n```\nor as a 4-bit Integer:\n```\n%OSCx.Message{address: \"/\", arguments: [%{time: 345827960115}]}\n```\nor to signal a time tag for 'processing immediately':\n```\n%OSCx.Message{address: \"/\", arguments: [%{time: :immediate}]}\n```\n\n#","ref":"examples.html#examples-with-time-tags"},{"type":"extras","title":"Example with a List or 'Array' in OSC: - Examples","doc":"```\n%OSCx.Message{address: \"/\", arguments: [1, 2, [:A, :B], 3, 4]}\n```\n\n#","ref":"examples.html#example-with-a-list-or-array-in-osc"},{"type":"extras","title":"Example with True and False: - Examples","doc":"```\n%OSCx.Message{address: \"/\", arguments: [true, false]}\n```\n\n#","ref":"examples.html#example-with-true-and-false"},{"type":"extras","title":"Example with Nil or Null: - Examples","doc":"You can use either Elixir's `nil` or the atom `:null` and both will be encoded to OSC's Null type:\n```\n%OSCx.Message{address: \"/\", arguments: [nil]}\n```\n\n#","ref":"examples.html#example-with-nil-or-null"},{"type":"extras","title":"Example with Impulse: - Examples","doc":"Impulse isalso known as Infinitum in OSC 1.0 Spec, or 'Bang':\n```\n%OSCx.Message{address: \"/\", arguments: [:impulse]}\n```","ref":"examples.html#example-with-impulse"},{"type":"extras","title":"Sending messages - Examples","doc":"Even though OSC messages are transport mechanism agnostic, they are commonly sent and received using UDP sockets.\n\nThe Erlang `:gen_utp` module can be used for this purpose. It provides the functions necessary for communicating with sockets using the UDP protocol.\n\n#","ref":"examples.html#sending-messages"},{"type":"extras","title":"Example - Examples","doc":"You can use [Protokol](https://hexler.net/protokol) to see the the output of this message.\n```\n# IP or host and port number for the UDP connection\nip_address = '127.0.0.1' # This could be changed to named address, like 'localhost'\nport_num = 8000 # In this example, this is the default port used by Protokol\n\n# Open a port\n{:ok, port} = :gen_udp.open(0, [:binary, {:active, true}])\n\n# Encode the message\nosc_message = %OSCx.Message{address: \"/greeting\", arguments: [\"Hello\", \"world\"]} |> OSCx.encode()\n\n# Send message\n:gen_udp.send(port, ip_address, port_num, osc_message)\n```\n\n![Screenshot of Protokol OSC inspection tool showing the message \"Hello, world\" being received.](assets/protokol-screenshot-greeting-example.png)","ref":"examples.html#example"},{"type":"extras","title":"Sending bundles - Examples","doc":"#","ref":"examples.html#sending-bundles"},{"type":"extras","title":"Example - Examples","doc":"You can use [Protokol](https://hexler.net/protokol) to see the the output of this message.\n```\n# IP or host and port number for the UDP connection\nip_address = '127.0.0.1' # This could be changed to named address, like 'localhost'\nport_num = 8000 # In this example, this is the default port used by Protokol\n\n# Open a port\n{:ok, port} = :gen_udp.open(0, [:binary, {:active, true}])\n\n# Encode the message\nosc_bundle =\n  %OSCx.Bundle{\n    time: %{seconds: 80, fraction: 2230576435},\n    elements: [\n      %OSCx.Message{address: \"/one\", arguments: []},\n      %OSCx.Message{address: \"/two\", arguments: []},\n      %OSCx.Message{address: \"/three\", arguments: []},\n    ]\n  } |> OSCx.encode()\n\n# Send message\n:gen_udp.send(port, ip_address, port_num, osc_bundle)\n```\n\n![Screenshot of Protokol OSC inspection tool showing the messages one, two and three from the bundle being received.](assets/protokol-screenshot-bundle-example.png)","ref":"examples.html#example"},{"type":"extras","title":"Receiving messages - Examples","doc":"Below is an example of a simple GenServer which can be used to send and receive OSC messages from [SuperCollider](https://supercollider.github.io/), an audio platform:\n\n```\ndefmodule SC do\n  use GenServer\n\n  @impl true\n  def init(_state) do\n    # Open a port and add the UDP socket to the state\n    {:ok, socket} = :gen_udp.open(0, [:binary, {:active, true}])\n\n    # In this example, the socket will be used as the state of the GenServer.\n    {:ok, socket}\n  end\n\n  @impl true\n  def handle_cast({:send, osc_bin_msg}, state) do\n    # This could be changed to named address, like 'localhost'\n    ip_address = ~c\"localhost\"\n    sc_port_num = 57110 # Default for SuperCollider\n    :gen_udp.send(state, ip_address, sc_port_num, osc_bin_msg)\n\n    {:noreply, state}\n  end\n\n  @impl true\n  def handle_info(msg, state) do\n    case msg do\n      {:udp, _process_port, _ip_addr, _port_num, res} ->\n        IO.inspect(res, label: \"Binary message received\")\n        IO.inspect(Message.decode(res), label: \"\\nDecoded message\")\n        \n      _ ->\n        # Do nothing\n    end\n\n    {:noreply, state}\n  end\n\n  def start_link() do\n    GenServer.start_link(SC, nil)\n  end\n\n  def send(pid, osc_bin_msg) do\n    GenServer.cast(pid, {:send, osc_bin_msg})\n  end\nend\n```\n\nLet's start up the GenServer, and use it to send an OSC message to SuperCollider:\n```\n{:ok, sc_pid} = SC.start_link()\n\nencoded_osc_msg =\n  %Message{address: \"/status\"}\n  |> OSCx.encode()\n\nSC.send(sc_pid, encoded_osc_msg)\n```\n\nIf successfull, you'll see something like the following, with the OSC message recieved from SuperCollider:\n\n```\nBinary message received:  >\n\nDecoded message: %OSCx.Message{\n  address: \"/status.reply\",\n  arguments: [1, 0, 0, 2, 111, 0.024587105959653854, 0.06639117002487183,\n   44100.0, 44099.94006284866]\n}\n```\n\nYou can explore the SuperCollider example further in the [OSCx tour Livebook](livebook/oscx_tour.livemd).","ref":"examples.html#receiving-messages"},{"type":"extras","title":"Arguments and types","doc":"# Arguments and types\n\nThe purpose of this page is to document:\n\n- What arguments are\n- Which types of arguments are recgonised by this library\n- How Elixir types are encoded to OSC\n- How OSC types are decoded to Elixir.","ref":"arguments_and_types.html"},{"type":"extras","title":"About OSC arguments - Arguments and types","doc":"Arguments represent the main data or 'payload' in an OSC message.\n\nArguments are typed. In an OSC message, the type of an argument is listed in the OSC tag type string, in the order the argument appears in the message.\n\n![OSC message diagram](assets/osc-message.png)\n\nOSC is encoded into 4-byte chunks. That is Ok for 32-bit data types are they are already of a 4-byte length, but for arbintary length data (like strings or blobs) padding with some null bytes may be needed to ensure each chunk is 4 bytes. \n\nThe following diagram shows how each chunk is kept at exactly 4 bytes length, by adding padding where necessary:\n```\nChunk 1: [D] [D] [D] [D]\nChunk 2: [D] [D] [D] [_]\nChunk 3: [D] [D] [_] [_]\nChunk 4: [D] [D] [D] [D]\n```\n*Key:* Where `[D]` is a byte of data and and `[_]` is a byte of empty padding.","ref":"arguments_and_types.html#about-osc-arguments"},{"type":"extras","title":"How Elixir types are encoded to OSC types - Arguments and types","doc":"From Elixir, arguments are added to an OSC message using the `%OSCx.Message{}` struct. For example, below is a message with an integer of `2`, a float of `440.5` and a string `\"phaser\"`:\n```\niex> %OSCx.Message{arguments: [2, 440.5, \"phaser\"]}\n%OSCx.Message{address: \"/\", arguments: [2, 440.5, \"phaser\"]}\n``` \nThe following table shows how these and other Elixir types are encoded to OSC types:\n\n| Elixir type      | Example argument value      | OSC type                                   | OSC spec version  |\n| ---------------- | --------------------------- | ------------------------------------------ | ----------------- |\n| Integer (32 bit) | `2`                         | 32-bit integer                             | 1.0+ required     |\n| Integer (64 bit) | `9_223_372_036_854_775_800` | 64-bit big-endian two’s complement integer | 1.0+ non-standard |\n| Float (32 bit)   | `440.5`                     | 32-bit float                               | 1.0+ required     |\n| Float (64 bit)   | `2.2250738585072014e-308`   | 64-bit float                               | 1.0+ non-standard |\n| String           | `\"phaser\"`                  | String                                     | 1.0+ required     |\n| Bitstring        | ` >`        | Blob                                       | 1.0+ required     |\n| Atom             | `:loud`                     | Symbol                                     | 1.0+ non-standard |\n| Map with `:seconds` and `:fraction` keys or `:time` key | `%{seconds: _, fraction: _ }` or `%{time: value}` | Time tag         | 1.1+ required     |\n| Map with `:midi` key | `%{midi: [153, 77, 63]}` | 4 byte MIDI message                        | 1.0+ non-standard |\n| Map with `:char` key | `%{char: 'A'}`            | ASCII char (32-bit)                        | 1.0+ non-standard |\n| Map with `:rgba` key with list of RGBA integers | `%{rgba: [255, 255, 90, 20]}` | 4-bit RGBA colour                          | 1.0+ non-standard |\n| List             | `[1, 2, 3]`                 | Array                                      | 1.0+ non-standard |\n\nNote that the:\n- MIDI map can accept values as Lists (e.g. `%{midi: [153, 77, 63]}`), Binary (e.g. `%{midi:  >}`) or other Elixir type that can be encoded to a 4-byte binary value, see: `OSCx.Encoder.midi/1`\n- The Char map can accept a value as a String (e.g. `%{char: \"A\"}`), Charlist (`%{char: ~c\"A\"}` or `%{char: 'A'}`) or Integer (e.g. `%{char: 65}`). See: `OSCx.Encoder.char/1`.\n- The Time tag map can be in the format of either:\n  - `%{seconds: seconds, fraction: fraction}` where `seconds` and `fraction` are 32-bit integers\n  - `%{time: value}` where `value` is a 64-bit integer.\n  - `%{time: :immediate}` to signal \"process immediately\"\n\n### 'Special' types\nThere are also some types that are encoded differently as they don't carry a variable value like those above:\n\n| Elixir value                | OSC type                                   | OSC spec version |\n| --------------------------- | ------------------------------------------ | ---------------- |\n| `true`                      | True                                       | 1.1+ required    |\n| `false`                     | False                                      | 1.1+ required    |\n| `nil`                       | Null                                       | 1.1+ required    |\n| `:null`                     | Null                                       | 1.1+ required    |\n| `:impulse`                  | Impulse (also known as Infinitum in OSC 1.0 Spec, or 'Bang')  | 1.1+ required    |\n\n![OSC message diagram](assets/osc-message-special.png)\n\nThese can be used in the same way by including them within the `arguments: []` list in the `%OSCx.Message{}` struct, for example:\n```\niex> iex> %OSCx.Message{arguments: [true]}\n%OSCx.Message{\n  address: \"/\",\n  arguments: [true],\n}\n\niex> iex> %OSCx.Message{arguments: [:impulse, 1.0, false, \"hello\"]}\n%OSCx.Message{\n  address: \"/\",\n  arguments: [:impulse, 1.0, false, \"hello\"]\n}\n```","ref":"arguments_and_types.html#how-elixir-types-are-encoded-to-osc-types"},{"type":"extras","title":"How OSC types are decoded to Elixir types - Arguments and types","doc":"#","ref":"arguments_and_types.html#how-osc-types-are-decoded-to-elixir-types"},{"type":"extras","title":"Decoding of arguments - Arguments and types","doc":"Decoding is simply the reverse of the above, where the following OSC type becomes the equivalent Elixir type:\n- Integer -> Integer\n- Float -> Float\n- String -> String\n- Blob -> Bitstring (binary)\n- Symbol -> Atom\n- MIDI -> `%{midi: value}`\n- Time tag -> `%{time: value, seconds: seconds, fraction: fraction}` where `value` is a 64-bit integer and `seconds` and `fraction` are 32-bit integers extracted from the value\n- ASCII Char -> `%{char: value}` where value will be a charlist\n- RGBA colour -> `%{rgba: [r, g, b, a]}` where there is an integer value for (r)ed, g(reen), b(lue) and a(lpha) colour channels.\n- Array -> List\n\n##","ref":"arguments_and_types.html#decoding-of-arguments"},{"type":"extras","title":"Symbols, Atoms and Strings - Arguments and types","doc":"By default, Symbols are converted to Elixir atoms. This behaviour can be changed to return a string instead. See `OSCx.Decoder.set_symbol_to_atom/1`.\n\n#","ref":"arguments_and_types.html#symbols-atoms-and-strings"},{"type":"extras","title":"Decoding of 'special' types - Arguments and types","doc":"The types of **True**, **False**, **Null**, **Impulse** will be decoded to `true`, `false`, `nil`, `:impulse` respectively and stored in `arguments: []` list of the `%OSCx.Message{}` struct, e.g.:\n```\n%OSCx.Message{\n  address: \"/\",\n  arguments: [],\n  tag_types: [nil]\n}\n```","ref":"arguments_and_types.html#decoding-of-special-types"},{"type":"extras","title":"More information - Arguments and types","doc":"- [README file](README.md)\n- [Time tags and synchronisation](time_tags.md)","ref":"arguments_and_types.html#more-information"},{"type":"extras","title":"Time tags and synchronisation","doc":"# Time tags and synchronisation\n\nThe purpose of this page is to document:\n\n- Summarise timing in OSC\n- Describe the role of the OSC time tag\n- Descibe how timing is used when processing OSC Bundles and Messages.\n\nIt draws heavily from the the Temporal Semantics and OSC Time Tags section of the [OpenSoundControl Specification 1.0](https://opensoundcontrol.stanford.edu/spec-1_0.html).","ref":"time_tags.html"},{"type":"extras","title":"Summary - Time tags and synchronisation","doc":"- OSC doesn't offer clock synchronization. It does offer a way to encode timing information though OSC time tags.\n\n- When an OSC packet has a single message, the server should process it immediately. If it's a bundle, the time tag within the bundle determines when to process the contained messages.\n\n- Time tags are represented as 64-bit fixed point numbers, with the first 32 bits indicating seconds since January 1, 1900, and the last 32 bits representing fractional seconds. To represent a time-tag, this library uses a map of `%{seconds: seconds, fraction: fraction}` where `seconds` and `fraction` are 32-bit integers, or `%{time: value}` where `value` is a 64-bit integer.\n\n- If a time tag has 63 zero bits followed by a one, it means \"immediately\".\n\n- Messages in the same bundle should be processed one after the other without interruption.\n\n- The order of method invocation is unspecified when an address pattern is dispatched to multiple methods. But messages within a bundle should be processed in the order they appear in the packet.\n\n- Bundles within bundles must have time tags greater than or equal to their enclosing bundle. The atomicity requirement for messages in the same bundle doesn't apply to bundles within a bundle.","ref":"time_tags.html#summary"},{"type":"extras","title":"Time management - Time tags and synchronisation","doc":"An OSC (Open Sound Control) server relies on access to accurate time data but it lacks built-in clock synchronization mechanisms. OSC servers need to work with precise time values for efficient operation. This is where OSC time tags come in.\n\nOSC time tags are represented as 64-bit fixed point numbers, where the:\n- first 32 bits specify the number of seconds elapsed since midnight on January 1, 1900\n- last 32 bits represent fractional parts of a second with high precision, approximately 200 picoseconds.\n\nTo signal immediate exection, a time tag can be created with a value comprising 63 zero bits followed by a one in the least significant bit. In this library you can use `%{time: :immediate}` for this purpose.\n\n#","ref":"time_tags.html#time-management"},{"type":"extras","title":"Formatting time in this library - Time tags and synchronisation","doc":"A time tag in this library is represented by a map. There are two different versions, depending on your needs:\n- Map with a 64-bit fixed point number:\n```\n%{time: value}\n```\n- Map with 32-bit integers representing seconds and fraction of a second:\n```\n%{seconds: seconds, fraction: fraction}\n```\n\nSee `OSCx.Encoder.time/1` and `OSCx.Decoder.time/2` for more information.","ref":"time_tags.html#formatting-time-in-this-library"},{"type":"extras","title":"Message processing - Time tags and synchronisation","doc":"When receiving an OSC packet, the behaviour differs based on its content:\n\n- ##","ref":"time_tags.html#message-processing"},{"type":"extras","title":"Single OSC Message - Time tags and synchronisation","doc":"If an incoming OSC packet contains only a single OSC message, the OSC server should promptly invoke the corresponding OSC method, processing it without delay.\n\n- ##","ref":"time_tags.html#single-osc-message"},{"type":"extras","title":"OSC Bundle - Time tags and synchronisation","doc":"When the packet contains an OSC bundle, the time tag associated with the bundle determines when the bundled OSC messages' corresponding methods should be invoked.\n  \n  If the time represented by the time tag is equal to or earlier than the current time, the OSC server should process the messages immediately. However, if the time tag indicates a future time, the OSC server must store the bundle until the specified time is reached and then execute the appropriate OSC methods.\n\n  - Atomic Message Processing: Messages within the same OSC bundle are treated as atomic units. Their corresponding OSC methods should be invoked consecutively, without any other processing occurring between these method invocations.\n\n  - Order of Method Invocation: When an OSC address pattern is dispatched to multiple OSC methods, the order in which these matching methods are invoked is not specified. However, when an OSC bundle contains multiple OSC messages, the methods corresponding to these messages must be invoked in the same order as the messages appear in the packet.\n\n    See [Order of invocation example](https://opensoundcontrol.stanford.edu/spec-1_0-examples.html#bundledispatchorder) in the OSC Spec.\n\n  - Nested Bundles: In the case of bundles within bundles, it's important to note that the OSC time tag of the enclosed bundle must be greater than or equal to the OSC time tag of the enclosing bundle. It's worth mentioning that the atomicity requirement for OSC messages in the same bundle does not apply to OSC bundles within an OSC bundle.","ref":"time_tags.html#osc-bundle"},{"type":"extras","title":"More information - Time tags and synchronisation","doc":"- [README file](README.md)\n- [Arguments and types](arguments_and_types.md)\n- Temporal Semantics and OSC Time Tags section of the [OpenSoundControl Specification 1.0](https://opensoundcontrol.stanford.edu/spec-1_0.html).","ref":"time_tags.html#more-information"},{"type":"extras","title":"OSCx tour","doc":"# OSCx tour\n\n```elixir\nMix.install([{:oscx, \"~> 0.1.1\"}])\n```","ref":"oscx_tour.html"},{"type":"extras","title":"Setup - OSCx tour","doc":"Create aliases for Message, Encoder and Decoder.\n\n```elixir\nalias OSCx.{Message, Encoder, Decoder}\n```\n\n \n\n```\n[OSCx.Message, OSCx.Encoder, OSCx.Decoder]\n```","ref":"oscx_tour.html#setup"},{"type":"extras","title":"Create an OSC message - OSCx tour","doc":"An OSC message is defined using the `%OSCx.Message{}` struct.\n\nYou can use the `OSCx.Message.new/1` function, e.g.:\n\n```\nMessage.new(address: \"/status\", arguments: [\"my string argument\"])\n```\n\nor populate the struct directly, see the cell below:\n\n```elixir\nmsg = %Message{address: \"/status\", arguments: []}\n```\n\n \n\n```\n%OSCx.Message{address: \"/status\", arguments: []}\n```","ref":"oscx_tour.html#create-an-osc-message"},{"type":"extras","title":"Encode the message - OSCx tour","doc":"The `OSCx.encode/1` function can be used to encode the struct to an OSC binary message.\n\n```elixir\nencoded_osc_msg = OSCx.encode(msg)\n```\n\n \n\n```\n >\n```\n\nThe message above is now ready to be sent via UDP.","ref":"oscx_tour.html#encode-the-message"},{"type":"extras","title":"Send an encoded message using UDP - OSCx tour","doc":"#","ref":"oscx_tour.html#send-an-encoded-message-using-udp"},{"type":"extras","title":"Download Protokol - OSCx tour","doc":"For this part, you may want to download an OSC data inspection tool, like [Protokol](https://hexler.net/protokol).\n\nProtokol is a free responsive console app for monitoring and logging control protocols, such as OSC. It can log MIDI and Gamepad inputs.\n\nIt's available on Mac, Windows and Linux.\n\nWe'll use Protokol to act as the OSC server (recipient) of our message.\n\n#","ref":"oscx_tour.html#download-protokol"},{"type":"extras","title":"Enable OSC in Protokol - OSCx tour","doc":"Once Protokol is installed, load it and click the 'OSC' tab. Make sure **Enabled** is checked on.\n\n#","ref":"oscx_tour.html#enable-osc-in-protokol"},{"type":"extras","title":"Open a UDP port - OSCx tour","doc":"Erlang and therefore Elixir comes with the `:gen_udp` library for opening UDP ports and sending and receiving messages.\n\nLets use it to send an OSC message to [Protokol](https://hexler.net/protokol). By default, Protokol listens on port `8000` but you can change this if necessary.\n\nAssuming Protokol is running on the same machine as this Livebook, we'll use the localhost IP address of `127.0.0.1` (or you can change this to `'localhost'` if you prefer).\n\n```elixir\nip_address = ~c\"127.0.0.1\"\nport_num = 8000\n\n# Open a port\n{:ok, port} = :gen_udp.open(0, [:binary, {:active, true}])\n```\n\n \n\n```\n{:ok, #Port }\n```\n\nNow the port is open, lets send our message using `:gen_udp.send/4` as below:\n\n```elixir\n:gen_udp.send(port, ip_address, port_num, encoded_osc_msg)\n```\n\n \n\n```\n:ok\n```\n\nYou should now see our message with `\"/status\"` on Protokol's OSC tab:\n![](files/protokol-screenshot-01.png)\n\n \n\nTo make it more interesting, lets send a more complex message by adding arguments to the `%OSCx.Messages{arguments: []}` key like this:\n\n```elixir\nencoded_osc_msg =\n  %Message{\n    address: \"/some/address\",\n    arguments: [1, 2.0, [:A, :B, :C], \"Hello world\", true, false, nil, :impulse]\n  }\n  |> OSCx.encode()\n```\n\n \n\n```\n >\n```\n\nThis message has a mix of OSC types and values:\n\n| OSC type | Elixir type                | Elixir example   |\n| -------- | -------------------------- | ---------------- |\n| Integer  | Integer                    | `1`              |\n| Float    | Float                      | `2.0`            |\n| Array    | List                       | `[1, 2, 3]`      |\n| String   | String                     | `\"Hello world\"`  |\n| Symbol   | Atom                       | `:A`             |\n| True     | true                       | `true`           |\n| False    | false                      | `false`          |\n| Null     | nil                        | `nil` or `:null` |\n| Impulse  | Atom (of value `:impulse`) | `:impulse`       |\n\nLet's send it as before:\n\n```elixir\n:gen_udp.send(port, ip_address, port_num, encoded_osc_msg)\n```\n\n \n\n```\n:ok\n```\n\nYou should now see our richer OSC message on the second line in Protokol:\n![](files/protokol-screenshot-02.png)\nUnfortunately Protokol doesn't show the array of symbols. This is something you may come across in OSC. Some types are optional parts of the OSC standard and are not implemented on all software.\n\nExamples of some optional or non-standard types are:\n\n* Symbols (e.g. Atoms in Elixir)\n* 64-bit integers or floats\n* Single ASCII character\n* RGBA colour\n* MIDI message\n* Array (e.g. List in Elixir)\n\nYou can send an OSC array of other data instead:\n\n```elixir\nbin_osc_msg =\n  %Message{\n    address: \"/some/address\",\n    arguments: [1, 2, 3, [\"X\", \"Y\", \"Z\"]]\n  }\n  |> OSCx.encode()\n\n:gen_udp.send(port, ip_address, port_num, bin_osc_msg)\n```\n\n \n\n```\n:ok\n```\n\nYou can see on the third line that the list values in the last was also received:\n![](files/protokol-screenshot-03.png)\n\n \n\nMIDI messages via OSC can also be sent by wrapping the binary MIDI message in a `%{midi: value}` map as follows:\n\n```elixir\nbin_osc_msg =\n  %Message{\n    address: \"/some/address\",\n    arguments: [%{midi:  >}]\n  }\n  |> OSCx.encode()\n\n:gen_udp.send(port, ip_address, port_num, bin_osc_msg)\n```\n\n \n\n```\n:ok\n```\n\nYou should see the MIDI message type and data appear in the Protokol screen:\n![](files/protokol-screenshot-04.png)\n\n \n\nThe OSC Spec requires a 4-byte MIDI message, defined as: port id, status byte, data1, and data2.\n\nHowever, MIDI messages are predominatly consist of 3-bytes (a status byte, followed by 2 data bytes). If a 3 byte MIDI message is given, OSCx will prepend a port ID of ` >` to make up the 1 byte difference.\n\nIf the OSC server (receiver) supports it (Protokol doesn't), RGBA colours can be sent as follows:\n\n```elixir\nbin_osc_msg =\n  %Message{\n    address: \"/some/address\",\n    arguments: [%{rgba: [153, 234, 69, 1]}]\n  }\n  |> OSCx.encode()\n\n:gen_udp.send(port, ip_address, port_num, bin_osc_msg)\n```\n\n \n\n```\n:ok\n```","ref":"oscx_tour.html#open-a-udp-port"},{"type":"extras","title":"Receiving OSC messages - OSCx tour","doc":"Using OSC to send and receive messages from audio software is a common use-case.\n\nThis example assumes SuperCollider is installed and running on the same machine as this Livebook.\n\n#","ref":"oscx_tour.html#receiving-osc-messages"},{"type":"extras","title":"Install SuperCollider - OSCx tour","doc":"In this example we'll use [SuperCollider](https://supercollider.github.io/). SuperCollider is an 'audio synthesis and algorithmic composition platform'. It's used by musicials, artists and researchers working with sound.\n\nWe can interact with SuperCollider using OSC to produce audio.\n\nSuperCollider is available on a range of platforms: https://supercollider.github.io/downloads\n\n#","ref":"oscx_tour.html#install-supercollider"},{"type":"extras","title":"Boot SuperCollider - OSCx tour","doc":"Launch SuperCollider, and from it's menu, select **Server > Boot Server** for it to boot.\n\n#","ref":"oscx_tour.html#boot-supercollider"},{"type":"extras","title":"Create a GenServer - OSCx tour","doc":"SuperCollider listents to UDP port `57110` by default. We'll use the GenServer's state to hold the UDP socket, and implement a `handle_info` callback where we'll receive UDP messages from SuperCollider.\n\n```elixir\ndefmodule SC do\n  use GenServer\n\n  @impl true\n  def init(_state) do\n    # Open a port and add the UDP socket to the state\n    {:ok, socket} = :gen_udp.open(0, [:binary, {:active, true}])\n\n    {:ok, socket}\n  end\n\n  @impl true\n  def handle_cast({:send, osc_bin_msg}, state) do\n    # This could be changed to named address, like 'localhost'\n    ip_address = ~c\"localhost\"\n    sc_port_num = 57110\n    :gen_udp.send(state, ip_address, sc_port_num, osc_bin_msg)\n\n    {:noreply, state}\n  end\n\n  @impl true\n  def handle_info(msg, state) do\n    case msg do\n      {:udp, _process_port, _ip_addr, _port_num, res} ->\n        IO.inspect(res, label: \"Binary message received\")\n        IO.inspect(Message.decode(res), label: \"\\nDecoded message\")\n        state\n\n      _ ->\n        state\n    end\n\n    {:noreply, state}\n  end\n\n  def start_link() do\n    GenServer.start_link(SC, nil)\n  end\n\n  def send(pid, osc_bin_msg) do\n    GenServer.cast(pid, {:send, osc_bin_msg})\n  end\nend\n```\n\n \n\n```\n{:module, SC,  >, {:send, 2}}\n```\n\nLet's start our GenServer so it's ready to send and receive messages:\n\n```elixir\n{:ok, sc_pid} = SC.start_link()\n```\n\n \n\n```\n{:ok, #PID }\n```\n\nLet's create a simple message and watch for a response.\n\nOne of the simplest messages to SuperCollider is sending a message with the address `\"/status\"` and no arguments:\n\n```elixir\nencoded_osc_msg =\n  %Message{address: \"/status\"}\n  |> OSCx.encode()\n\nSC.send(sc_pid, encoded_osc_msg)\n```\n\n \n\n```\n:ok\n```\n\n \n\n```\nBinary message received:  >\n\nDecoded message: %OSCx.Message{\n  address: \"/status.reply\",\n  arguments: [1, 0, 0, 2, 111, 0.024587105959653854, 0.06639117002487183,\n   44100.0, 44099.94006284866]\n}\n```\n\nIf successfull, you'll see from `IO.inspect` the binary OSC message received e.g. ` >` and it decoded as a struct:\n\n```\n%OSCx.Message{\n  address: \"/status.reply\",\n  arguments: [1, 0, 0, 2, 111, 0.029667392373085022, 0.14714068174362183]\n}\n```\n\n###","ref":"oscx_tour.html#create-a-genserver"},{"type":"extras","title":"Address - OSCx tour","doc":"Notice how SuperCollider sends the same address back as we sent but with \".reply\" appended to it: `\"/status.reply\"`. When sending commands to SuperCollider, replies will often follow this convention.\n\n###","ref":"oscx_tour.html#address"},{"type":"extras","title":"Arguments - OSCx tour","doc":"The arguments have meaning in SuperCollider, but you can see a mixture of integers and floats were returned.\n\nLet's try another simple message, this time we'll send a message with the address `\"/version\"` and no arguments:\n\n```elixir\nencoded_osc_msg =\n  %Message{address: \"/version\"}\n  |> OSCx.encode()\n\nSC.send(sc_pid, encoded_osc_msg)\n```\n\n \n\n```\n:ok\n```\n\n \n\n```\nBinary message received:  >\n\nDecoded message: %OSCx.Message{\n  address: \"/version.reply\",\n  arguments: [\"scsynth\", 3, 13, \".0\", \"Version-3.13.0\", \"3188503\"]\n}\n```\n\nIn this case, you can see version information sent back in the decoded message arguments:\n\n```\n%OSCx.Message{\n  address: \"/version.reply\",\n  arguments: [\"scsynth\", 3, 13, \".0\", \"Version-3.13.0\", \"3188503\"]\n}\n```\n\nYou message may be slightly different to this, depending on what you've installed locally.\n\n##","ref":"oscx_tour.html#arguments"},{"type":"extras","title":"Play a sound - OSCx tour","doc":"You can also try asking SuperCollider to play some audio:\n\n```elixir\n# Send a message to play the default sound with a new synth\nsynth_definition_name = \"default\"\nsynth_node_id = 400\nadd_action = 0\nadd_target_id = 0\n\nencoded_osc_msg =\n  %Message{\n    address: \"/s_new\",\n    arguments: [synth_definition_name, synth_node_id, add_action, add_target_id]\n  }\n  |> OSCx.encode()\n\nSC.send(sc_pid, encoded_osc_msg)\n```\n\n \n\n```\n:ok\n```\n\nSuperCollider's default tone should be playing.\n\n##","ref":"oscx_tour.html#play-a-sound"},{"type":"extras","title":"Stop the sound - OSCx tour","doc":"To stop the sound, evaluate the next cell:\n\n```elixir\nencoded_osc_msg =\n  %Message{\n    address: \"/n_free\",\n    arguments: [synth_node_id]\n  }\n  |> OSCx.encode()\n\nSC.send(sc_pid, encoded_osc_msg)\n```\n\n \n\n```\n:ok\n```","ref":"oscx_tour.html#stop-the-sound"},{"type":"extras","title":"Wrap up - OSCx tour","doc":"This ends the simple tour of the OSCx library.\n\nIn summary:\n\n* Create OSCx messages by using the `%OSCx.Message{}` struct\n* Encode them with `OSCx.encode/1`\n* You can send and receive OSC messages via UDP, using Erlang's `:gen_utp` library.\n* When receiving messages, as with our SuperCollider example above, use the `OSCx.decode/1` function and it will parse the binary OSC message and populate the `%OSCx.Message{}` struct for you.","ref":"oscx_tour.html#wrap-up"},{"type":"extras","title":"License (MIT)","doc":"MIT License\n\nCopyright (c) 2023 David Haubenschild.\n\nPermission is hereby granted, free of charge, to any person obtaining a copy\nof this software and associated documentation files (the \"Software\"), to deal\nin the Software without restriction, including without limitation the rights\nto use, copy, modify, merge, publish, distribute, sublicense, and/or sell\ncopies of the Software, and to permit persons to whom the Software is\nfurnished to do so, subject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in all\ncopies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\nIMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\nFITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\nAUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\nLIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\nOUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE\nSOFTWARE.","ref":"license.html"}],"content_type":"text/markdown"}